############################################################
# -*- coding: utf-8 -*-
#
#       #   #  #   #   #  ####
#      ##  ##  #  ##  #     #
#     # # # #  # # # #     ###
#    #  ##  #  ##  ##        #
#   #   #   #  #   #     ####
#
# Python-based Tool for interaction with the 10micron mounts
# GUI with PyQT5 for python
# Python  v3.6.4
#
# Michael Würtenberger
# (c) 2016, 2017, 2018
#
# Licence APL2.0
#
###########################################################
import logging
import time
import PyQt5
import queue
import math
import copy
from mount import mount_command
from mount import mount_statusfast
from mount import mount_statusmedium
from mount import mount_statusslow
from mount import mount_statusonce
from mount import mount_getalignmodel
from mount import mount_setalignmodel
from mount import mount_getmodelnames
from mount import mount_modelhandling
from analyse import analysedata
from baseclasses import checkIP
from astrometry import transform


class MountDispatcher(PyQt5.QtCore.QThread):
    logger = logging.getLogger(__name__)

    # needed signals for mount connections
    signalMountConnected = PyQt5.QtCore.pyqtSignal(dict)
    signalCancelRunTargetRMS = PyQt5.QtCore.pyqtSignal()

    # signals for data transfer to other threads
    signalMountAzAltPointer = PyQt5.QtCore.pyqtSignal(float, float)
    signalMountLimits = PyQt5.QtCore.pyqtSignal()
    signalAlignmentStars = PyQt5.QtCore.pyqtSignal()
    signalMountShowAlignmentModel = PyQt5.QtCore.pyqtSignal()
    signalMountShowModelNames = PyQt5.QtCore.pyqtSignal()
    signalSlewFinished = PyQt5.QtCore.pyqtSignal()

    CYCLE = 200
    signalDestruct = PyQt5.QtCore.pyqtSignal()

    statusReference = {
        '0': 'Tracking',
        '1': 'Stopped after STOP',
        '2': 'Slewing to park position',
        '3': 'Unparking',
        '4': 'Slewing to home position',
        '5': 'Parked',
        '6': 'Slewing or going to stop',
        '7': 'Tracking Off no move',
        '8': 'Motor low temperature',
        '9': 'Tracking outside limits',
        '10': 'Following Satellite',
        '11': 'User OK Needed',
        '98': 'Unknown Status',
        '99': 'Error'
    }

    data = {
        # positions of greenwich at start
        'SiteLatitude': '51:28:37',
        'SiteLongitude': '00:00:00',
        'SiteHeight': '46',
        'MountIP': '',
        'MountMAC': '',
        'MountPort': 3490,
        'LocalSiderealTime': '',
        # date of 01.05.2018
        'JulianDate': '2458240',
        'FW': 0
    }

    mountStatus = {
            'Fast': False,
            'Medium': False,
            'Slow': False,
            'Once': False,
            'GetAlign': False,
            'SetAlign': False,
            'GetName': False,
            'Command': False
    }

    def __init__(self, app, thread):
        super().__init__()
        self.app = app
        self.thread = thread
        self.isRunning = False
        self.mutexIsRunning = PyQt5.QtCore.QMutex()
        self.mutexIPChange = PyQt5.QtCore.QMutex()
        self.commandDispatcherQueue = queue.Queue()
        self.cycleTimer = None
        # getting all supporting classes assigned
        self.mountModelHandling = mount_modelhandling.MountModelHandling(self.app, self.data)
        self.analyse = analysedata.Analyse(self.app)
        self.transform = transform.Transform(self.app)
        self.checkIP = checkIP.CheckIP()

        # getting all threads setup
        # commands sending thread
        self.threadMountCommandRunner = PyQt5.QtCore.QThread()
        self.workerMountCommandRunner = mount_command.MountCommandRunner(self.app, self.threadMountCommandRunner, self.data, self.signalMountConnected, self.mountStatus)
        self.threadMountCommandRunner.setObjectName("MountCommandRunner")
        self.workerMountCommandRunner.moveToThread(self.threadMountCommandRunner)
        self.threadMountCommandRunner.started.connect(self.workerMountCommandRunner.run)
        # fast status thread
        self.threadMountStatusRunnerFast = PyQt5.QtCore.QThread()
        self.workerMountStatusRunnerFast = mount_statusfast.MountStatusRunnerFast(self.app, self.threadMountStatusRunnerFast, self.data, self.signalMountConnected, self.mountStatus)
        self.threadMountStatusRunnerFast.setObjectName("MountStatusRunnerFast")
        self.workerMountStatusRunnerFast.moveToThread(self.threadMountStatusRunnerFast)
        self.threadMountStatusRunnerFast.started.connect(self.workerMountStatusRunnerFast.run)
        # medium status thread
        self.threadMountStatusRunnerMedium = PyQt5.QtCore.QThread()
        self.workerMountStatusRunnerMedium = mount_statusmedium.MountStatusRunnerMedium(self.app, self.threadMountStatusRunnerMedium, self.data, self.signalMountConnected, self.mountStatus)
        self.threadMountStatusRunnerMedium.setObjectName("MountStatusRunnerMedium")
        self.workerMountStatusRunnerMedium.moveToThread(self.threadMountStatusRunnerMedium)
        self.threadMountStatusRunnerMedium.started.connect(self.workerMountStatusRunnerMedium.run)
        # slow status thread
        self.threadMountStatusRunnerSlow = PyQt5.QtCore.QThread()
        self.workerMountStatusRunnerSlow = mount_statusslow.MountStatusRunnerSlow(self.app, self.threadMountStatusRunnerSlow, self.data, self.signalMountConnected, self.mountStatus)
        self.threadMountStatusRunnerSlow.setObjectName("MountStatusRunnerSlow")
        self.workerMountStatusRunnerSlow.moveToThread(self.threadMountStatusRunnerSlow)
        self.threadMountStatusRunnerSlow.started.connect(self.workerMountStatusRunnerSlow.run)
        # once status thread
        self.threadMountStatusRunnerOnce = PyQt5.QtCore.QThread()
        self.workerMountStatusRunnerOnce = mount_statusonce.MountStatusRunnerOnce(self.app, self.threadMountStatusRunnerOnce, self.data, self.signalMountConnected, self.mountStatus)
        self.threadMountStatusRunnerOnce.setObjectName("MountStatusRunnerOnce")
        self.workerMountStatusRunnerOnce.moveToThread(self.threadMountStatusRunnerOnce)
        self.threadMountStatusRunnerOnce.started.connect(self.workerMountStatusRunnerOnce.run)
        # get alignment model
        self.threadMountGetAlignmentModel = PyQt5.QtCore.QThread()
        self.workerMountGetAlignmentModel = mount_getalignmodel.MountGetAlignmentModel(self.app, self.threadMountGetAlignmentModel, self.data, self.signalMountConnected, self.mountStatus)
        self.threadMountGetAlignmentModel.setObjectName("MountGetAlignmentModel")
        self.workerMountGetAlignmentModel.moveToThread(self.threadMountGetAlignmentModel)
        self.threadMountGetAlignmentModel.started.connect(self.workerMountGetAlignmentModel.run)
        # set alignment model
        self.threadMountSetAlignmentModel = PyQt5.QtCore.QThread()
        self.workerMountSetAlignmentModel = mount_setalignmodel.MountSetAlignmentModel(self.app, self.threadMountSetAlignmentModel, self.data, self.signalMountConnected, self.mountStatus)
        self.threadMountSetAlignmentModel.setObjectName("MountSetAlignmentModel")
        self.workerMountSetAlignmentModel.moveToThread(self.threadMountSetAlignmentModel)
        self.threadMountSetAlignmentModel.started.connect(self.workerMountSetAlignmentModel.run)
        # get model names
        self.threadMountGetModelNames = PyQt5.QtCore.QThread()
        self.workerMountGetModelNames = mount_getmodelnames.MountGetModelNames(self.app, self.threadMountGetModelNames, self.data, self.signalMountConnected, self.mountStatus)
        self.threadMountGetModelNames.setObjectName("MountGetModelNames")
        self.workerMountGetModelNames.moveToThread(self.threadMountGetModelNames)
        self.threadMountGetModelNames.started.connect(self.workerMountGetModelNames.run)

        self.cancelRunTargetRMS = False
        self.runTargetRMS = False
        self.programAlignmentModelStatus = None
        self.commandDispatch = {
            'RunTargetRMSAlignment':
                {
                    'Worker': [
                        {
                            'Button': self.app.ui.btn_runTargetRMSAlignment,
                            'Method': self.runTargetRMSAlignment,
                            'Cancel': self.app.ui.btn_cancelRunTargetRMSAlignment
                        }
                    ]
                },
            'ClearAlign':
                {
                    'Worker': [
                        {
                            'Button': self.app.ui.btn_clearModel,
                            'Method': self.mountModelHandling.clearAlign
                        }
                    ]
                },
            'DeleteWorstPoint':
                {
                    'Worker': [
                        {
                            'Button': self.app.ui.btn_deleteWorstPoint,
                            'Method': self.deleteWorstPoint,
                        }
                    ]
                },
            'ReloadAlignmentModel':
                {
                    'Worker': [
                        {
                            'Button': self.app.ui.btn_reloadAlignmentModel,
                            'Method': self.reloadAlignmentModel,
                        }
                    ]
                },
            'SetRefractionParameter':
                {
                    'Worker': [
                        {
                            'Button': self.app.ui.btn_setRefractionParameters,
                            'Method': self.workerMountStatusRunnerMedium.getStatusMedium,
                        }
                    ]
                },
            'FLIP':
                {
                    'Worker': [
                        {
                            'Button': self.app.ui.btn_flipMount,
                            'Method': self.flipMount,
                        }
                    ]
                },
            'Shutdown':
                {
                    'Worker': [
                        {
                            'Button': self.app.ui.btn_mountShutdown,
                            'Method': self.mountShutdown,
                        }
                    ]
                }
        }
        # signal slot
        self.app.ui.btn_setRefractionParameters.clicked.connect(lambda: self.commandDispatcherQueue.put('SetRefractionParameter'))
        self.app.ui.btn_runTargetRMSAlignment.clicked.connect(lambda: self.commandDispatcherQueue.put('RunTargetRMSAlignment'))
        self.app.ui.btn_deleteWorstPoint.clicked.connect(lambda: self.commandDispatcherQueue.put('DeleteWorstPoint'))
        self.app.ui.btn_flipMount.clicked.connect(lambda: self.commandDispatcherQueue.put('FLIP'))
        self.app.ui.btn_reloadAlignmentModel.clicked.connect(lambda: self.commandDispatcherQueue.put('ReloadAlignmentModel'))
        self.app.ui.btn_mountShutdown.clicked.connect(lambda: self.commandDispatcherQueue.put('Shutdown'))
        self.app.ui.btn_clearModel.clicked.connect(lambda: self.commandDispatcherQueue.put('ClearAlign'))

        self.signalCancelRunTargetRMS.connect(self.setCancelRunTargetRMS)
        self.app.ui.btn_saveModel.clicked.connect(self.saveSelectedModel)
        self.app.ui.btn_loadModel.clicked.connect(self.loadSelectedModel)
        self.app.ui.btn_deleteModel.clicked.connect(self.deleteSelectedModel)
        self.app.ui.listModelName.itemDoubleClicked.connect(self.getListAction)
        self.signalMountShowModelNames.connect(self.setModelNamesList)
        self.signalMountConnected.connect(self.setMountConnectionStatus)

    def initConfig(self):
        try:
            if 'MountIP' in self.app.config:
                self.app.ui.le_mountIP.setText(self.app.config['MountIP'])
                self.app.sharedMountDataLock.lockForWrite()
                self.data['MountIP'] = self.app.config['MountIP']
                self.app.sharedMountDataLock.unlock()
            if 'MountMAC' in self.app.config:
                self.app.ui.le_mountMAC.setText(self.app.config['MountMAC'])
                self.app.sharedMountDataLock.lockForWrite()
                self.data['MountMAC'] = self.app.config['MountMAC']
                self.app.sharedMountDataLock.unlock()
            if 'CheckAutoRefractionContinuous' in self.app.config:
                self.app.ui.checkAutoRefractionContinous.setChecked(self.app.config['CheckAutoRefractionContinuous'])
            if 'CheckAutoRefractionNotTracking' in self.app.config:
                self.app.ui.checkAutoRefractionNotTracking.setChecked(self.app.config['CheckAutoRefractionNotTracking'])
            if 'CheckAutoRefractionNone' in self.app.config:
                self.app.ui.checkAutoRefractionNone.setChecked(self.app.config['CheckAutoRefractionNone'])
            # if we have already stored the site data, we use it until we get new information from mount
            if 'SiteLongitude' in self.app.config:
                self.data['SiteLongitude'] = copy.copy(self.app.config['SiteLongitude'])
            if 'SiteLatitude' in self.app.config:
                self.data['SiteLatitude'] = copy.copy(self.app.config['SiteLatitude'])
            if 'SiteHeight' in self.app.config:
                self.data['SiteHeight'] = copy.copy(self.app.config['SiteHeight'])
                self.app.signalMountSiteData.emit(self.data['SiteLatitude'],
                                                  self.data['SiteLongitude'],
                                                  self.data['SiteHeight'])
        except Exception as e:
            self.logger.error('item in config.cfg not be initialize, error:{0}'.format(e))
        finally:
            pass

    def storeConfig(self):
        self.app.config['MountIP'] = self.app.ui.le_mountIP.text()
        self.app.config['MountMAC'] = self.app.ui.le_mountMAC.text()
        self.app.config['CheckAutoRefractionContinuous'] = self.app.ui.checkAutoRefractionContinous.isChecked()
        self.app.config['CheckAutoRefractionNotTracking'] = self.app.ui.checkAutoRefractionNotTracking.isChecked()
        self.app.config['CheckAutoRefractionNone'] = self.app.ui.checkAutoRefractionNone.isChecked()
        # if we had a connection to the mount, the site data should be there.
        if self.mountStatus['Once']:
            self.app.config['SiteLongitude'] = copy.copy(self.data['SiteLongitude'])
            self.app.config['SiteLatitude'] = copy.copy(self.data['SiteLatitude'])
            self.app.config['SiteHeight'] = copy.copy(self.data['SiteHeight'])

    def setCancelRunTargetRMS(self):
        self.cancelRunTargetRMS = True

    def changedSettings(self):
        self.mutexIPChange.lock()
        # stopping all interaction
        if self.isRunning:
            # stopping thread for chang of parameters
            self.logger.info('Stopping threads for IP change')
            self.workerMountStatusRunnerFast.stop()
            self.workerMountStatusRunnerMedium.stop()
            self.workerMountStatusRunnerSlow.stop()
            self.workerMountStatusRunnerOnce.stop()
            self.workerMountGetAlignmentModel.stop()
            self.workerMountSetAlignmentModel.stop()
            self.workerMountCommandRunner.stop()
            self.workerMountGetModelNames.stop()
            self.app.sharedMountDataLock.lockForWrite()
            self.data['MountIP'] = self.app.ui.le_mountIP.text()
            self.data['MountMAC'] = self.app.ui.le_mountMAC.text()
            self.logger.info('Setting IP address for mount to: {0}'.format(self.data['MountIP']))
            self.app.sharedMountDataLock.unlock()
            # and restarting for using new parameters
            self.threadMountCommandRunner.start()
            self.threadMountGetModelNames.start()
            self.threadMountSetAlignmentModel.start()
            self.threadMountGetAlignmentModel.start()
            self.threadMountStatusRunnerOnce.start()
            self.threadMountStatusRunnerSlow.start()
            self.threadMountStatusRunnerMedium.start()
            self.threadMountStatusRunnerFast.start()
        else:
            self.logger.info('IP change when threads not running')
            self.app.sharedMountDataLock.lockForWrite()
            self.data['MountIP'] = self.app.ui.le_mountIP.text()
            self.data['MountMAC'] = self.app.ui.le_mountMAC.text()
            self.app.sharedMountDataLock.unlock()
        self.app.messageQueue.put('Setting IP address for mount to: {0}\n'.format(self.data['MountIP']))
        self.mutexIPChange.unlock()

    def run(self):
        self.app.ui.le_mountIP.editingFinished.connect(self.changedSettings, type=PyQt5.QtCore.Qt.QueuedConnection)
        self.logger.info('mount dispatcher started')
        # sending default status to gui in red
        self.app.signalSetMountStatus.emit(0)
        self.mutexIsRunning.lock()
        if not self.isRunning:
            self.isRunning = True
        self.mutexIsRunning.unlock()
        self.threadMountCommandRunner.start()
        self.threadMountSetAlignmentModel.start()
        self.threadMountGetModelNames.start()
        self.threadMountGetAlignmentModel.start()
        self.threadMountStatusRunnerOnce.start()
        self.threadMountStatusRunnerSlow.start()
        self.threadMountStatusRunnerMedium.start()
        self.threadMountStatusRunnerFast.start()
        self.signalDestruct.connect(self.destruct, type=PyQt5.QtCore.Qt.BlockingQueuedConnection)
        self.cycleTimer = PyQt5.QtCore.QTimer(self)
        self.cycleTimer.setSingleShot(False)
        self.cycleTimer.timeout.connect(self.doCommand)
        self.cycleTimer.start(self.CYCLE)

    def stop(self):
        self.mutexIsRunning.lock()
        if self.isRunning:
            self.isRunning = False
            self.signalDestruct.emit()
            self.thread.quit()
            self.thread.wait()
        self.mutexIsRunning.unlock()
        # stopping all interaction
        self.logger.info('mount dispatcher stopped')

    @PyQt5.QtCore.pyqtSlot()
    def destruct(self):
        self.cycleTimer.stop()
        self.workerMountStatusRunnerFast.stop()
        self.workerMountStatusRunnerMedium.stop()
        self.workerMountStatusRunnerSlow.stop()
        self.workerMountStatusRunnerOnce.stop()
        self.workerMountGetAlignmentModel.stop()
        self.workerMountSetAlignmentModel.stop()
        self.workerMountGetModelNames.stop()
        self.workerMountCommandRunner.stop()
        self.signalDestruct.disconnect(self.destruct)
        self.app.ui.le_mountIP.editingFinished.disconnect(self.changedSettings)

    def doCommand(self):
        if not self.commandDispatcherQueue.empty():
            command = self.commandDispatcherQueue.get()
            if isinstance(command, dict):
                # transferring complete working set
                self.manualCommandDispatcher(command)
            else:
                # doing standard work based on init
                self.commandDispatcher(command)

    def manualCommandDispatcher(self, command):
        # running through all necessary commands
        for work in command['Worker']:
            # if we want to color a button, which one
            if 'Button' in work:
                self.app.signalChangeStylesheet.emit(work['Button'], 'running', True)
            if 'Parameter' in work:
                parameter = []
                for p in work['Parameter']:
                    parameter.append(p)
                work['Method'](*parameter)
            else:
                work['Method']()
            time.sleep(0.2)
            if 'Button' in work:
                self.app.signalChangeStylesheet.emit(work['Button'], 'running', False)
            if 'Cancel' in work:
                self.app.signalChangeStylesheet.emit(work['Cancel'], 'cancel', False)

    def commandDispatcher(self, command):
        # if we have a command in dispatcher
        if command in self.commandDispatch:
            # running through all necessary commands
            for work in self.commandDispatch[command]['Worker']:
                # if we want to color a button, which one
                if 'Button' in work:
                    self.app.signalChangeStylesheet.emit(work['Button'], 'running', True)
                if 'Parameter' in work:
                    parameter = []
                    for p in work['Parameter']:
                        parameter.append(p)
                    work['Method'](*parameter)
                else:
                    work['Method']()
                time.sleep(0.2)
                if 'Button' in work:
                    self.app.signalChangeStylesheet.emit(work['Button'], 'running', False)
                if 'Cancel' in work:
                    self.app.signalChangeStylesheet.emit(work['Cancel'], 'cancel', False)

    def setMountConnectionStatus(self, status):
        for key in status:
            self.mountStatus[key] = status[key]
        stat = 0
        for key in self.mountStatus:
            if self.mountStatus[key]:
                stat += 1
        if stat == 0:
            # red
            self.app.signalSetMountStatus.emit(0)
        elif stat == len(self.mountStatus):
            # all connected green
            self.app.signalSetMountStatus.emit(2)
        else:
            # otherwise yellow
            self.app.signalSetMountStatus.emit(1)

    def setModelNamesList(self):
        self.app.ui.listModelName.clear()
        for name in self.data['ModelNames']:
            self.app.ui.listModelName.addItem(name)
        self.app.ui.listModelName.sortItems()
        self.app.ui.listModelName.update()

    def getListAction(self):
        name = self.app.ui.listModelName.currentItem().text()
        question = 'Action with mount model:\n\n\t{0}\n\n'.format(name)
        value = self.app.dialogMessageLoadSaveDelete(self.app, 'Mount model management', question)
        if value == 0:
            action = {
                'Worker': [
                    {
                        'Button': self.app.ui.btn_loadModel,
                        'Parameter': [name],
                        'Method': self.mountModelHandling.loadModel,
                    }
                ]
            }
            self.commandDispatcherQueue.put(action)
        elif value == 1:
            action = {
                'Worker': [
                    {
                        'Button': self.app.ui.btn_saveModel,
                        'Parameter': [name],
                        'Method': self.mountModelHandling.saveModel,
                    }
                ]
            }
            self.commandDispatcherQueue.put(action)
        elif value == 2:
            action = {
                'Worker': [
                    {
                        'Button': self.app.ui.btn_deleteModel,
                        'Parameter': [name],
                        'Method': self.mountModelHandling.deleteModel,
                    }
                ]
            }
            self.commandDispatcherQueue.put(action)
        else:
            pass

    def saveSelectedModel(self):
        if self.app.ui.listModelName.currentItem() is not None:
            name = self.app.ui.listModelName.currentItem().text()
        else:
            name = ''
        name, ok = self.app.dialogInputText(self.app, 'Please enter the model name', 'Model name:', name)
        if ok:
            # limit length of name to 15 characters
            name = name[:15]
            action = {
                'Worker': [
                    {
                        'Button': self.app.ui.btn_saveModel,
                        'Parameter': [name],
                        'Method': self.mountModelHandling.saveModel,
                    }
                ]
            }
            self.commandDispatcherQueue.put(action)

    def loadSelectedModel(self):
        if self.app.ui.listModelName.currentItem() is not None:
            name = self.app.ui.listModelName.currentItem().text()
            action = {
                'Worker': [
                    {
                        'Button': self.app.ui.btn_loadModel,
                        'Parameter': [name],
                        'Method': self.mountModelHandling.loadModel,
                    }
                ]
            }
            self.commandDispatcherQueue.put(action)

    def deleteSelectedModel(self):
        if self.app.ui.listModelName.currentItem() is not None:
            name = self.app.ui.listModelName.currentItem().text()
            action = {
                'Worker': [
                    {
                        'Button': self.app.ui.btn_deleteModel,
                        'Parameter': [name],
                        'Method': self.mountModelHandling.deleteModel,
                    }
                ]
            }
            self.commandDispatcherQueue.put(action)

    def mountShutdown(self):
        # mount has to run
        if self.workerMountCommandRunner.socket.state() != PyQt5.QtNetwork.QAbstractSocket.ConnectedState:
            return
        commandSet = {'command': ':shutdown#', 'reply': ''}
        self.app.mountCommandQueue.put(commandSet)
        while len(commandSet['reply']) == 0:
            time.sleep(0.1)
        if commandSet['reply'] == '1':
            self.workerMountCommandRunner.connected = False
            time.sleep(1)
            self.logger.info('Shutdown mount manually')
            self.app.messageQueue.put('Shutting mount down !\n')
        else:
            self.logger.error('error: {0}'.format(commandSet['reply']))
            self.app.messageQueue.put('#BRError in mount shutdown\n')

    def flipMount(self):
        commandSet = {'command': ':FLIP#', 'reply': ''}
        self.app.mountCommandQueue.put(commandSet)
        while len(commandSet['reply']) == 0:
            time.sleep(0.1)
        if commandSet['reply'] == '0':
            self.app.messageQueue.put('#BRFlip Mount could not be executed\n')
            self.logger.error('error: {0}'.format(commandSet['reply']))

    def syncMountModel(self, ra, dec):
        self.logger.info('ra:{0} dec:{1}'.format(ra, dec))
        self.app.mountCommandQueue.put(':Sr{0}#'.format(ra))
        self.app.mountCommandQueue.put(':Sd{0}#'.format(dec))
        self.app.mountCommandQueue.put(':CMCFG0#')
        # send sync command
        commandSet = {'command': ':CM#', 'reply': ''}
        self.app.mountCommandQueue.put(commandSet)
        while len(commandSet['reply']) == 0:
            time.sleep(0.1)
        if commandSet['reply'][:5] == 'Coord':
            self.logger.info('Mount modeling synced')
            return True
        else:
            self.logger.warning('Error in sync mount modeling')
            return False

    def programBatchData(self, data):
        if not('RaJNow' in data and 'DecJNow' in data):
            self.logger.warning('RaJNow or DecJNow not in data file')
            self.messageQueue.put('Mount coordinates missing\n')
            return
        if not('RaJNowSolved' in data and 'DecJNowSolved' in data):
            self.logger.warning('RaJNowSolved or DecJNowSolved not in data file')
            self.messageQueue.put('Solved data missing\n')
            return
        if not('Pierside' in data and 'LocalSiderealTimeFloat' in data):
            self.logger.warning('Pierside and LocalSiderealTimeFloat not in data file')
            self.messageQueue.put('Time and Pierside missing\n')
            return
        self.app.messageQueue.put('#BWProgramming alignment model data\n')
        self.workerMountSetAlignmentModel.result = None
        self.workerMountSetAlignmentModel.setAlignmentModel(data)
        while self.workerMountSetAlignmentModel.result is None:
            time.sleep(0.1)
            PyQt5.QtWidgets.QApplication.processEvents()
        if self.workerMountSetAlignmentModel.result:
            self.logger.info('Model successful finished!')
            self.app.messageQueue.put('#BWProgrammed alignment model with {0} points\n'.format(len(data['Index'])))
        else:
            self.logger.warning('Model could not be calculated with current data!')
            self.app.messageQueue.put('#BRProgramming alignment model finished with errors\n')
        self.commandDispatcherQueue.put('ReloadAlignmentModel')

    def runTargetRMSAlignment(self):
        self.runTargetRMS = True
        self.cancelRunTargetRMS = False
        self.app.messageQueue.put('#BWTarget RMS Run started\n')
        self.app.sharedMountDataLock.lockForRead()
        condition = ('Number' not in self.data or self.data['Number'] < 4)
        self.app.sharedMountDataLock.unlock()
        if condition:
            self.runTargetRMS = False
            return
        while True:
            self.app.sharedMountDataLock.lockForRead()
            data = self.data['RMS']
            self.app.sharedMountDataLock.unlock()
            if self.cancelRunTargetRMS:
                break
            if data < float(self.app.ui.targetRMS.value()):
                break
            if self.deleteWorstPoint():
                break
        if self.cancelRunTargetRMS:
            self.app.messageQueue.put('#BRTarget RMS Run canceled\n')
        else:
            self.app.messageQueue.put('#BWTarget RMS Run finished\n')
        self.runTargetRMS = False

    def reloadAlignmentModel(self):
        self.workerMountGetAlignmentModel.getAlignmentModel()
        # wait form alignment model to be downloaded
        while True:
            self.app.sharedMountDataLock.lockForRead()
            condition = not self.data['ModelLoading']
            self.app.sharedMountDataLock.unlock()
            if condition:
                break
            time.sleep(0.2)

    def deleteWorstPoint(self):
        # if there are less than 4 point, optimization can't take place
        self.app.sharedMountDataLock.lockForRead()
        if self.data['Number'] < 4:
            return True
        # find worst point
        maxError = 0
        worstPointIndex = 0
        for i in range(0, self.data['Number']):
            if self.data['ModelError'][i] > maxError:
                worstPointIndex = i
                maxError = self.data['ModelError'][i]
        self.app.messageQueue.put('Deleting worst point  {0:02d} with AZ:  {1:05.1f}  ALT:  {2:04.1f}  and error of:  {3:05.1f}\n'
                                  .format(worstPointIndex + 1,
                                          self.data['ModelAzimuth'][worstPointIndex],
                                          self.data['ModelAltitude'][worstPointIndex],
                                          maxError))
        self.app.sharedMountDataLock.unlock()
        commandSet = {'command': ':delalst{0:d}#'.format(worstPointIndex + 1), 'reply': ''}
        self.app.mountCommandQueue.put(commandSet)
        while len(commandSet['reply']) == 0:
            time.sleep(0.1)
        time.sleep(0.2)
        if commandSet['reply'] == '1':
            # point could be deleted, feedback from mount ok
            self.logger.info('Deleting worst point {0} with error of:  {1}'.format(worstPointIndex+1, maxError))
            # get new calculated alignment model from mount
            self.app.messageQueue.put('\tPoint deleted\n')
        else:
            self.app.messageQueue.put('#BR\tPoint could not be deleted \n')
            self.logger.warning('Point {0} could not be deleted'.format(worstPointIndex))
        self.workerMountGetAlignmentModel.getAlignmentModel()
        # wait form alignment model to be downloaded
        while True:
            self.app.sharedMountDataLock.lockForRead()
            condition = not self.data['ModelLoading']
            self.app.sharedMountDataLock.unlock()
            if condition:
                break
            time.sleep(0.2)

    def retrofitMountData(self, modelingData):
        self.app.sharedMountDataLock.lockForRead()
        self.app.sharedModelingDataLock.lockForWrite()
        if len(self.data['ModelError']) == len(modelingData['Index']):
            modelingData['ModelErrorOptimized'] = list()
            modelingData['RaErrorOptimized'] = list()
            modelingData['DecErrorOptimized'] = list()
            for i in range(0, len(self.data['ModelError'])):
                modelingData['ModelErrorOptimized'].append(self.data['ModelError'][i])
                modelingData['RaErrorOptimized'].append(self.data['ModelError'][i] * math.sin(math.radians(self.data['ModelErrorAngle'][i])))
                modelingData['DecErrorOptimized'].append(self.data['ModelError'][i] * math.cos(math.radians(self.data['ModelErrorAngle'][i])))
            self.app.messageQueue.put('Data synced\n')
            returnValue = True
        else:
            self.logger.warning('Size mount modeling {0} and modeling data {1} do not fit !'.format(len(modelingData['Index']), len(self.data['ModelError'])))
            self.app.messageQueue.put('Mount Model and Model Data could not be synced\n')
            self.app.messageQueue.put('Error data sync mismatch!\n')
            returnValue = False
        self.app.sharedMountDataLock.unlock()
        self.app.sharedModelingDataLock.unlock()
        return returnValue



############################################################
# -*- coding: utf-8 -*-
#
#       #   #  #   #   #  ####
#      ##  ##  #  ##  #     #
#     # # # #  # # # #     ###
#    #  ##  #  ##  ##        #
#   #   #   #  #   #     ####
#
# Python-based Tool for interaction with the 10micron mounts
# GUI with PyQT5 for python
# Python  v3.6.4
#
# Michael Würtenberger
# (c) 2016, 2017, 2018
#
# Licence APL2.0
#
###########################################################
import logging
import os
import PyQt5
import time
import copy
import operator
import numpy
from astrometry import transform


class ModelPoints:
    logger = logging.getLogger(__name__)

    def __init__(self, app):
        self.app = app
        self.transform = transform.Transform(self.app)
        self.horizonPoints = list()
        self.modelPoints = list()
        self.celestialEquator = list()
        # signal slot
        self.app.ui.btn_loadInitialModelPoints.clicked.connect(self.selectInitialModelPointsFileName)
        self.app.ui.btn_saveInitialModelPoints.clicked.connect(self.saveInitialModelPoints)
        self.app.ui.btn_saveInitialModelPointsAs.clicked.connect(self.saveInitialModelPointsAs)
        self.app.ui.btn_loadFullModelPoints.clicked.connect(self.selectFullModelPointsFileName)
        self.app.ui.btn_saveFullModelPoints.clicked.connect(self.saveFullModelPoints)
        self.app.ui.btn_saveFullModelPointsAs.clicked.connect(self.saveFullModelPointsAs)
        self.app.ui.btn_loadHorizonMask.clicked.connect(self.selectHorizonPointsFileName)
        self.app.ui.btn_saveHorizonMask.clicked.connect(self.saveHorizonMask)
        self.app.ui.btn_saveHorizonMaskAs.clicked.connect(self.saveHorizonMaskAs)
        self.app.signalMountSiteData.connect(self.generateCelestialEquator)

    def initConfig(self):
        try:
            if 'HorizonPointsFileName' in self.app.config:
                self.app.ui.le_horizonPointsFileName.setText(self.app.config['HorizonPointsFileName'])
            if 'CheckUseMinimumHorizonLine' in self.app.config:
                self.app.ui.checkUseMinimumHorizonLine.setChecked(self.app.config['CheckUseMinimumHorizonLine'])
            if 'CheckUseFileHorizonLine' in self.app.config:
                self.app.ui.checkUseFileHorizonLine.setChecked(self.app.config['CheckUseFileHorizonLine'])
            if 'AltitudeMinimumHorizon' in self.app.config:
                self.app.ui.altitudeMinimumHorizon.setValue(self.app.config['AltitudeMinimumHorizon'])
            if 'ModelInitialPointsFileName' in self.app.config:
                self.app.ui.le_modelInitialPointsFileName.setText(self.app.config['ModelInitialPointsFileName'])
            if 'ModelFullPointsFileName' in self.app.config:
                self.app.ui.le_modelFullPointsFileName.setText(self.app.config['ModelFullPointsFileName'])
            if 'HorizonPointsFileName' in self.app.config and 'CheckUseMinimumHorizonLine' in self.app.config and 'CheckUseFileHorizonLine' in self.app.config and 'AltitudeMinimumHorizon' in self.app.config:
                self.loadHorizonPoints(self.app.config['HorizonPointsFileName'],
                                       self.app.config['CheckUseFileHorizonLine'],
                                       self.app.config['CheckUseMinimumHorizonLine'],
                                       self.app.config['AltitudeMinimumHorizon'])

        except Exception as e:
            self.logger.error('item in config.cfg could not be initialize, error:{0}'.format(e))
        finally:
            pass

    def storeConfig(self):
        self.app.config['HorizonPointsFileName'] = self.app.ui.le_horizonPointsFileName.text()
        self.app.config['CheckUseMinimumHorizonLine'] = self.app.ui.checkUseMinimumHorizonLine.isChecked()
        self.app.config['CheckUseFileHorizonLine'] = self.app.ui.checkUseFileHorizonLine.isChecked()
        self.app.config['AltitudeMinimumHorizon'] = self.app.ui.altitudeMinimumHorizon.value()
        self.app.config['ModelInitialPointsFileName'] = self.app.ui.le_modelInitialPointsFileName.text()
        self.app.config['ModelFullPointsFileName'] = self.app.ui.le_modelFullPointsFileName.text()

    def saveHorizonMask(self):
        filepath = os.getcwd() + '/config/' + self.app.ui.le_horizonPointsFileName.text()
        self.saveHorizonPoints(filepath)

    def saveHorizonMaskAs(self):
        value, ext = self.app.selectFile(self.app, 'Save horizon mask points file', '/config', 'Model point files (*.txt)', False)
        if value != '':
            self.app.ui.le_horizonPointsFileName.setText(os.path.basename(value))
            self.saveHorizonPoints(value)
        else:
            self.logger.warning('No model points file selected')

    def selectHorizonPointsFileName(self):
        value, ext = self.app.selectFile(self.app, 'Open horizon mask file', '/config', 'Horizon mask files (*.txt)', True)
        if value != '':
            self.app.ui.le_horizonPointsFileName.setText(os.path.basename(value))
            self.app.hemisphereWindow.selectHorizonPointsMode()
            self.app.hemisphereWindow.drawHemisphere()

    def saveModelPoints(self, modelPointsFileName):
        msg = None
        fileHandle = None
        if modelPointsFileName.strip() == '':
            msg = 'No Model Points Filename given!'
            self.logger.warning('No Model Points Filename given!')
            return msg
        try:
            fileHandle = open(modelPointsFileName + '.txt', 'w')
            for i in range(0, len(self.modelPoints)):
                fileHandle.write('MW-3:{0:03.2f}:{1:03.2f}\n'.format(self.modelPoints[i][0], self.modelPoints[i][1]))
            fileHandle.close()
        except Exception as e:
            msg = 'Error saving modeling points to file [{0}] error: {1}!'.format(modelPointsFileName, e)
            self.logger.warning('Error loading modeling points to file [{0}] error: {1}!'.format(modelPointsFileName, e))
        finally:
            if fileHandle:
                fileHandle.close()
            return msg

    def saveInitialModelPoints(self):
        filepath = os.getcwd() + '/config/' + self.app.ui.le_modelInitialPointsFileName.text()
        self.saveModelPoints(filepath)

    def saveInitialModelPointsAs(self):
        value, ext = self.app.selectFile(self.app, 'Save initial model points file', '/config', 'Model point files (*.txt)', False)
        if value != '':
            self.app.ui.le_modelInitialPointsFileName.setText(os.path.basename(value))
            self.saveModelPoints(value)
        else:
            self.logger.warning('No model points file selected')

    def selectInitialModelPointsFileName(self):
        value, ext = self.app.selectFile(self.app, 'Open initial model points file', '/config', 'Model points files (*.txt)', True)
        if value != '':
            value = os.path.basename(value)
            self.app.ui.le_modelInitialPointsFileName.setText(value)
            self.showInitialPoints(value)
        else:
            self.logger.warning('No file selected')

    def saveFullModelPoints(self):
        filepath = os.getcwd() + '/config/' + self.app.ui.le_modelFullPointsFileName.text()
        self.saveModelPoints(filepath)

    def saveFullModelPointsAs(self):
        value, ext = self.app.selectFile(self.app, 'Save full model points file', '/config', 'Model point files (*.txt)', False)
        if value != '':
            self.app.ui.le_modelFullPointsFileName.setText(os.path.basename(value))
            self.saveModelPoints(value)
        else:
            self.logger.warning('No model points file selected')

    def selectFullModelPointsFileName(self):
        value, ext = self.app.selectFile(self.app, 'Open full model points file', '/config', 'Model points files (*.txt)', True)
        if value != '':
            value = os.path.basename(value)
            self.app.ui.le_modelFullPointsFileName.setText(value)
            self.showFullPoints(value, self.app.ui.checkDeletePointsHorizonMask.isChecked(), self.app.ui.checkSortPoints.isChecked())
        else:
            self.logger.warning('No file selected')

    def loadModelPoints(self, modelPointsFileName, modeltype):
        p = []
        number = 0
        msg = None
        if modelPointsFileName.strip() == '':
            msg = 'No model points filename given!'
            self.logger.warning('No model points filename given!')
            return p, msg
        try:
            with open('config/' + modelPointsFileName + '.txt', 'r') as fileHandle:
                for line in fileHandle:
                    if line.startswith('GRID'):
                        # if grid, then its a TSX file (the sky x)
                        convertedLine = line.rstrip('\n').split()
                        point = (float(convertedLine[2]), float(convertedLine[3]))
                        number += 1
                        if modeltype == 'Refinement' and number > 3:
                            p.append(point)
                        elif modeltype == 'Base' and number <= 3:
                            p.append(point)
                    elif line.startswith('MW-3'):
                        # if mountwizzard3, it's native version 3
                        convertedLine = line.rstrip('\n').split(':')
                        p.append((float(convertedLine[1]), float(convertedLine[2])))
                    else:
                        # format is same as Per's Model Maker
                        convertedLine = line.rstrip('\n').split(':')
                        point = (int(convertedLine[0]), int(convertedLine[1]))
                        if len(convertedLine) == 2 and modeltype == 'Full':
                            p.append(point)
                        elif len(convertedLine) != 2 and modeltype == 'Initial':
                            p.append(point)
        except Exception as e:
            msg = 'Error loading modeling points from file [{0}] error: {1}!'.format(modelPointsFileName, e)
            self.logger.warning('Error loading modeling points from file [{0}] error: {1}!'.format(modelPointsFileName, e))
        finally:
            return p, msg

    def sortPoints(self):
        if len(self.modelPoints) == 0:
            self.logger.warning('There are no points to sort')
            return
        westSide = []
        eastSide = []
        a = sorted(self.modelPoints, key=operator.itemgetter(0))
        for i in range(0, len(a)):
            if a[i][0] >= 180:
                westSide.append((a[i][0], a[i][1]))
            else:
                eastSide.append((a[i][0], a[i][1]))
        westSide = sorted(westSide, key=operator.itemgetter(1))
        eastSide = sorted(eastSide, key=operator.itemgetter(1))
        self.modelPoints = westSide + eastSide

    def loadHorizonPoints(self, horizonPointsFileName, horizonByFile, horizonByAltitude, altitudeMinimumHorizon):
        self.horizonPoints = []
        if not (horizonByFile or horizonByAltitude):
            return
        hp = []
        msg = None
        if horizonByFile:
            if horizonPointsFileName == '':
                msg = 'No horizon points filename given !'
                return msg
            if not os.path.isfile(os.getcwd() + '/config/' + horizonPointsFileName + '.txt'):
                msg = 'Horizon points file does not exist !'
                self.logger.warning('Horizon points file does not exist')
            else:
                try:
                    with open(os.getcwd() + '/config/' + horizonPointsFileName + '.txt') as f:
                        for line in f:
                            if ':' in line:
                                # model maker format
                                m = line.rstrip('\n').split(':')
                            else:
                                # carte du ciel / skychart format
                                m = line.rstrip('\n').split(' ')
                            point = (int(m[0]), int(m[1]))
                            hp.append(point)
                    f.close()
                except Exception as e:
                    msg = 'Error loading horizon points: {0}'.format(e)
                    self.logger.error('Error loading horizon points: {0}'.format(e))
                    return msg
            hp = sorted(hp, key=operator.itemgetter(0))
        if len(hp) == 0:
            hp = ((0, 0), (360, 0))
        x = [i[0] for i in hp]
        y = [i[1] for i in hp]
        if horizonByAltitude:
            y = numpy.clip(y, altitudeMinimumHorizon, None)
        self.horizonPoints = [list(a) for a in zip(x, y)]
        return msg

    def saveHorizonPoints(self, horizonPointsFileName):
        msg = None
        fileHandle = None
        if horizonPointsFileName.strip() == '':
            msg = 'No horizon points filename given!'
            self.logger.warning('No Model Points Filename given!')
            return msg
        try:
            fileHandle = open(horizonPointsFileName + '.txt', 'w')
            for i in range(0, len(self.horizonPoints)):
                # saving in model maker format
                fileHandle.write('{0:03d}:{1:03d}\n'.format(int(self.horizonPoints[i][0]), int(int(self.horizonPoints[i][1]))))
            fileHandle.close()
        except Exception as e:
            msg = 'Error saving horizon points to file [{0}] error: {1}!'.format(horizonPointsFileName, e)
            self.logger.warning('Error loading horizon points to file [{0}] error: {1}!'.format(horizonPointsFileName, e))
        finally:
            if fileHandle:
                fileHandle.close()
        return msg

    def isAboveHorizonLine(self, point):
        x = range(0, 361)
        y = numpy.interp(x, [i[0] for i in self.horizonPoints], [i[1] for i in self.horizonPoints], left=None, right=None, period=None)
        if point[1] > y[int(point[0])]:
            return True
        else:
            return False

    def deleteBelowHorizonLine(self):
        i = 0
        while i < len(self.modelPoints):
            if self.isAboveHorizonLine(self.modelPoints[i]):
                i += 1
            else:
                del self.modelPoints[i]

    def deletePoints(self):
        self.modelPoints = list()
        self.app.workerModelingDispatcher.signalModelPointsRedraw.emit()

    def showInitialPoints(self, filename):
        self.modelPoints, msg = self.loadModelPoints(filename, 'Initial')
        self.app.messageQueue.put('ToModel>{0:02d}'.format(len(self.modelPoints)))
        self.app.workerModelingDispatcher.signalModelPointsRedraw.emit()

    def showFullPoints(self, filename, limitByHorizonMask, doSortingPoints):
        self.modelPoints, msg = self.loadModelPoints(filename, 'Full')
        if limitByHorizonMask:
            self.deleteBelowHorizonLine()
        if doSortingPoints:
            self.sortPoints()
        self.app.messageQueue.put('ToModel>{0:02d}'.format(len(self.modelPoints)))
        self.app.workerModelingDispatcher.signalModelPointsRedraw.emit()

    def generateDSOPoints(self, limitByHorizonMask, hoursPathLength, numberOfPathPoints, hoursPathLengthPreview):
        # we have no position of the mount -> therefore we can't calculate the path
        if 'RaJNow' not in self.app.workerMountDispatcher.data:
            return
        self.modelPoints = list()
        ra = copy.copy(self.app.workerMountDispatcher.data['RaJNow'])
        dec = copy.copy(self.app.workerMountDispatcher.data['DecJNow'])
        for i in range(0, numberOfPathPoints):
            ra = ra - float(i) * hoursPathLength / numberOfPathPoints - hoursPathLengthPreview
            az, alt = self.transform.transformERFA(ra, dec, 1)
            if alt > 0:
                self.modelPoints.append((az, alt))
        if limitByHorizonMask:
            self.deleteBelowHorizonLine()
        self.app.messageQueue.put('ToModel>{0:02d}'.format(len(self.modelPoints)))
        self.app.workerModelingDispatcher.signalModelPointsRedraw.emit()

    def generateMaxPoints(self, limitByHorizonMask, doSortingPoints):
        west = []
        east = []
        off = -5
        i = 0
        for dec in range(-15, 90, 10):
            if dec < 30:
                step = 10
            elif dec < 70:
                step = 10
            else:
                step = 30
            if i % 2:
                for ha in range(120 + off, -120 + off, -step):
                    az, alt = self.transform.topocentricToAzAlt(ha / 10, dec)
                    if alt > 0:
                        if az > 180:
                            east.insert(0, (az, alt))
                        else:
                            west.append((az, alt))
            else:
                for ha in range(-120 + off, 120 + off, step):
                    az, alt = self.transform.topocentricToAzAlt(ha / 10, dec)
                    if alt > 0:
                        if az > 180:
                            east.insert(0, (az, alt))
                        else:
                            west.append((az, alt))
            i += 1
        self.modelPoints = west + east
        if limitByHorizonMask:
            self.deleteBelowHorizonLine()
        if doSortingPoints:
            self.sortPoints()
        self.app.messageQueue.put('ToModel>{0:02d}'.format(len(self.modelPoints)))
        self.app.workerModelingDispatcher.signalModelPointsRedraw.emit()

    def generateNormalPoints(self, limitByHorizonMask, doSortingPoints):
        west = []
        east = []
        off = -5
        i = 0
        for dec in range(-15, 90, 15):
            if dec < 60:
                step = 10
            else:
                step = 20
            if i % 2:
                for ha in range(120 + off, -120 + off, -step):
                    az, alt = self.transform.topocentricToAzAlt(ha / 10, dec)
                    if alt > 0:
                        if az > 180:
                            east.insert(0, (az, alt))
                        else:
                            west.append((az, alt))
            else:
                for ha in range(-120 + off, 120 + off, step):
                    az, alt = self.transform.topocentricToAzAlt(ha / 10, dec)
                    if alt > 0:
                        if az > 180:
                            east.insert(0, (az, alt))
                        else:
                            west.append((az, alt))
            i += 1
        self.modelPoints = west + east
        if limitByHorizonMask:
            self.deleteBelowHorizonLine()
        if doSortingPoints:
            self.sortPoints()
        self.app.messageQueue.put('ToModel>{0:02d}'.format(len(self.modelPoints)))
        self.app.workerModelingDispatcher.signalModelPointsRedraw.emit()

    def generateMinPoints(self, limitByHorizonMask, doSortingPoints):
        west = list()
        east = list()
        off = -5
        i = 0
        for dec in range(-15, 90, 15):
            if dec < 60:
                step = 15
            else:
                step = 30
            if i % 2:
                for ha in range(120 + off, -120 + off, -step):
                    az, alt = self.transform.topocentricToAzAlt(ha / 10, dec)
                    if alt > 0:
                        if az > 180:
                            east.insert(0, (az, alt))
                        else:
                            west.append((az, alt))
            else:
                for ha in range(-120 + off, 120 + off, step):
                    az, alt = self.transform.topocentricToAzAlt(ha / 10, dec)
                    if alt > 0:
                        if az > 180:
                            east.insert(0, (az, alt))
                        else:
                            west.append((az, alt))
            i += 1
        self.modelPoints = west + east
        if limitByHorizonMask:
            self.deleteBelowHorizonLine()
        if doSortingPoints:
            self.sortPoints()
        self.app.messageQueue.put('ToModel>{0:02d}'.format(len(self.modelPoints)))
        self.app.workerModelingDispatcher.signalModelPointsRedraw.emit()

    def generateGridPoints(self, limitByHorizonMask, doSortingPoints, numberOfRows, numberOfColumns, altitudeMin, altitudeMax):
        west = list()
        east = list()
        i = 0
        for alt in range(altitudeMin, altitudeMax + 1, int((altitudeMax - altitudeMin) / (numberOfRows - 1))):
            if i % 2:
                for az in range(365 - int(360 / numberOfColumns), 0, -int(360 / numberOfColumns)):
                    if alt > 0:
                        if az > 180:
                            east.insert(0, (az, alt))
                        else:
                            west.append((az, alt))
            else:
                for az in range(5, 360, int(360 / numberOfColumns)):
                    if alt > 0:
                        if az > 180:
                            east.insert(0, (az, alt))
                        else:
                            west.append((az, alt))
            i += 1
        self.modelPoints = west + east
        if limitByHorizonMask:
            self.deleteBelowHorizonLine()
        if doSortingPoints:
            self.sortPoints()
        self.app.messageQueue.put('ToModel>{0:02d}'.format(len(self.modelPoints)))
        self.app.workerModelingDispatcher.signalModelPointsRedraw.emit()

    def generateInitialPoints(self, azimuth, altitude, numberOfPoints):
        self.modelPoints = list()
        for i in range(0, numberOfPoints):
            azp = i * 360 / numberOfPoints + azimuth
            if azp > 360:
                azp -= 360
            azp = int(azp)
            point = (azp, altitude)
            self.modelPoints.append(point)
        self.app.messageQueue.put('ToModel>{0:02d}'.format(len(self.modelPoints)))
        self.app.workerModelingDispatcher.signalModelPointsRedraw.emit()

    def generateCelestialEquator(self):
        self.celestialEquator = list()
        off = -5
        for dec in range(-15, 90, 15):
            for ha in range(120 + off, -120 + off, -2):
                az, alt = self.transform.topocentricToAzAlt(ha / 10, dec)
                if alt > 0:
                    self.celestialEquator.append((az, alt))



############################################################
# -*- coding: utf-8 -*-
#
# Python-based Tool for interaction with the 10micron mounts
# GUI with PyQT5 for python
# Python  v3.5
#
# Michael Würtenberger
# (c) 2016, 2017, 2018
#
# Licence APL2.0
#
############################################################
import json
import math
import os
from logging import getLogger


class Analyse:
    logger = getLogger(__name__)

    UPDATE = {'index': 'Index',
              'azimuth': 'Azimuth',
              'altitude': 'Altitude',
              'modelError': 'ModelError',
              'raError': 'RaError',
              'decError': 'DecError',
              'ra_Jnow': 'RaJNow',
              'dec_Jnow': 'DecJNow',
              'ra_sol_Jnow': 'RaJNowSolved',
              'dec_sol_Jnow': 'DecJNowSolved',
              'pierside': 'Pierside',
              'sidereal_time_float': 'LocalSiderealTimeFloat'}

    def __init__(self, app):
        self.filepath = '/analysedata'
        self.app = app

        self.app.ui.btn_split.clicked.connect(self.splitData)

    def splitData(self):
        mainFilename = self.app.ui.le_analyseFileName.text()
        data = self.loadDataRaw(mainFilename)
        lengthData = len(data['Index'])
        for i in range(3, lengthData):
            dataSplit = {}
            for key in data:
                dataSplit[key] = data[key][:i]
            splitFilename = mainFilename + '_split_{0:02d}'.format(i)
            self.saveData(dataSplit, splitFilename)

    def saveData(self, dataProcess, name):
        filenameData = os.getcwd() + self.filepath + '/' + name + '.dat'
        try:
            outfile = open(filenameData, 'w')
            json.dump(dataProcess, outfile)
            outfile.close()
        except Exception as e:
            self.logger.error('analyse data file {0}, Error : {1}'.format(filenameData, e))
            return

    def processTheSkyXLine(self, line):
        ra_sol = self.app.mount.degStringToDecimal(line[0:13], ' ')
        dec_sol = self.app.mount.degStringToDecimal(line[15:28], ' ')
        ra = self.app.mount.degStringToDecimal(line[30:43], ' ')
        dec = self.app.mount.degStringToDecimal(line[45:58], ' ')
        lst = self.app.mount.degStringToDecimal(line[61:70], ' ')
        return ra, dec, ra_sol, dec_sol, lst

    def loadTheSkyXData(self, filename):
        resultData = {}
        try:
            with open(filename) as infile:
                lines = infile.read().splitlines()
            infile.close()
            # site_latitude = self.app.mount.degStringToDecimal(lines[4][0:9], ' ')
            for i in range(5, len(lines)):
                ra, dec, ra_sol, dec_sol, lst = self.processTheSkyXLine(lines[i])
                if 'RaJ2000' in resultData:
                    resultData['RaJ2000'].append(ra)
                else:
                    resultData['RaJ2000'] = [ra]
                if 'DecJ2000' in resultData:
                    resultData['DecJ2000'].append(dec)
                else:
                    resultData['DecJ2000'] = [dec]
                ra_Jnow, dec_Jnow = self.app.mount.transformERFA(ra, dec, 3)
                if 'RaJNow' in resultData:
                    resultData['RaJNow'].append(ra_Jnow)
                else:
                    resultData['RaJNow'] = [ra_Jnow]
                if 'DecJNow' in resultData:
                    resultData['decJNow'].append(dec_Jnow)
                else:
                    resultData['DecJNow'] = [dec_Jnow]
                if 'LocalSiderealTimeFloat' in resultData:
                    resultData['LocalSiderealTimeFloat'].append(lst)
                else:
                    resultData['LocalSiderealTimeFloat'] = [lst]
                if 'LocalSiderealTime' in resultData:
                    resultData['LocalSiderealTime'].append(self.app.mount.decimalToDegree(lst, False, True))
                else:
                    resultData['LocalSiderealTime'] = [self.app.mount.decimalToDegree(lst, False, True)]
                if 'RaJ2000Solved' in resultData:
                    resultData['RaJ2000Solved'].append(ra_sol)
                else:
                    resultData['RaJ2000Solved'] = [ra_sol]
                if 'DecJ2000Solved' in resultData:
                    resultData['DecJ2000Solved'].append(dec_sol)
                else:
                    resultData['DecJ2000Solved'] = [dec_sol]
                ra_sol_Jnow, dec_sol_Jnow = self.app.mount.transformERFA(ra_sol, dec_sol, 3)
                if 'RaJNowSolved' in resultData:
                    resultData['RaJNowSolved'].append(ra_sol_Jnow)
                else:
                    resultData['RaJNowSolved'] = [ra_sol_Jnow]
                if 'DecJNowSolved' in resultData:
                    resultData['DecJNowSolved'].append(dec_sol_Jnow)
                else:
                    resultData['DecJNowSolved'] = [dec_sol_Jnow]
                ha = ra - lst
                az, alt = self.app.mount.transformERFA(ha, dec, 3)
                if 'Azimuth' in resultData:
                    resultData['Azimuth'].append(az)
                else:
                    resultData['Azimuth'] = [az]
                if 'Altitude' in resultData:
                    resultData['Altitude'].append(alt)
                else:
                    resultData['Altitude'] = [alt]
                if az <= 180:
                    pierside = 'E'
                else:
                    pierside = 'W'
                if 'Pierside' in resultData:
                    resultData['Pierside'].append(pierside)
                else:
                    resultData['Pierside'] = [pierside]
                if 'Index' in resultData:
                    resultData['Index'].append(i - 5)
                else:
                    resultData['index'] = [i - 5]
                if 'RaError' in resultData:
                    resultData['RaError'].append((ra - ra_sol) * 3600)
                else:
                    resultData['RaError'] = [(ra - ra_sol) * 3600]
                if 'DecError' in resultData:
                    resultData['DecError'].append((dec - dec_sol) * 3600)
                else:
                    resultData['DecError'] = [(dec - dec_sol) * 3600]
                if 'ModelError' in resultData:
                    resultData['ModelError'].append(math.sqrt((ra - ra_sol) * 3600 * (ra - ra_sol) * 3600 + (dec - dec_sol) * 3600 * (dec - dec_sol) * 3600))
                else:
                    resultData['ModelError'] = [math.sqrt((ra - ra_sol) * 3600 * (ra - ra_sol) * 3600 + (dec - dec_sol) * 3600 * (dec - dec_sol) * 3600)]
        except Exception as e:
            self.logger.error('error processing file {0}, Error : {1}'.format(filename, e))
            return {}
        return resultData

    def loadMountWizzardData(self, filename):
        try:
            infile = open(filename, 'r')
            dataJson = json.load(infile)
            infile.close()
        except Exception as e:
            self.logger.error('analyse data file {0}, Error : {1}'.format(filename, e))
            return {}
        # check if old file format
        if isinstance(dataJson, list):
            resultData = dict()
            for timestepdict in dataJson:
                for (keyData, valueData) in timestepdict.items():
                    if keyData in self.UPDATE:
                        keyData = self.UPDATE[keyData]
                    if keyData in resultData:
                        resultData[keyData].append(valueData)
                    else:
                        resultData[keyData] = [valueData]
        else:
            resultData = dataJson
        return resultData

    def loadDataRaw(self, filename):
        filenameData = os.getcwd() + self.filepath + '/' + filename + '.dat'
        if os.path.isfile(filenameData):
            infile = open(filenameData, 'r')
            dataJson = json.load(infile)
            infile.close()
            return dataJson
        else:
            return None

    def loadData(self, filename):
        filenameData = os.getcwd() + self.filepath + '/' + filename + '.dat'
        if os.path.isfile(filenameData):
            infile = open(filenameData, 'r')
            check = infile.read(8)
            infile.close()
            if check == '!TheSkyX':
                data = self.loadTheSkyXData(filenameData)
            else:
                data = self.loadMountWizzardData(filenameData)
            return data
        else:
            return {}


if __name__ == "__main__":
    logger = getLogger(__name__)
    from mount.mount_dispatcher import Mount
    filename = '10micron_model.dat'
    m = Mount
    a = Analyse(m)
    data = a.loadData(filename)
    # print(data)



############################################################
# -*- coding: utf-8 -*-
#
#       #   #  #   #   #  ####
#      ##  ##  #  ##  #     #
#     # # # #  # # # #     ###
#    #  ##  #  ##  ##        #
#   #   #   #  #   #     ####
#
# Python-based Tool for interaction with the 10micron mounts
# GUI with PyQT5 for python
# Python  v3.6.4
#
# Michael Würtenberger
# (c) 2016, 2017, 2018
#
# Licence APL2.0
#
###########################################################
import json
import logging
import platform
import socket
import timeit
import time
import PyQt5


class TheSkyX:
    logger = logging.getLogger(__name__)

    host = '127.0.0.1'
    port = 3040

    CAMERASTATUS = {'Not Connected': 'DISCONNECTED', 'Downloading Light': 'DOWNLOAD', 'Exposure complete': 'IDLE', 'Ready': 'IDLE', 'Exposing Light': 'INTEGRATING'}

    def __init__(self, main, app, data):
        # make main sources available
        self.main = main
        self.app = app
        self.data = data
        self.cancel = False
        self.checkIP = checkIP.CheckIP()
        self.mutexCancel = PyQt5.QtCore.QMutex()

        self.commandQueue = commandQueue
        self.isRunning = False
        self.mutexIsRunning = PyQt5.QtCore.QMutex()

        self.application = dict()
        self.application['Available'] = False
        self.application['Name'] = ''
        self.application['InstallPath'] = ''
        self.application['Status'] = ''
        self.application['Runtime'] = 'TheSkyX.exe'

        self.responseSuccess = '|No error. Error = 0.'

        if platform.system() == 'Windows':
            # sgpro only supported on local machine
            self.application['Available'], self.application['Name'], self.application['InstallPath'] = self.app.checkRegistrationKeys('Sequence Generator')
            if self.application['Available']:
                self.app.messageQueue.put('Found Imaging: {0}\n'.format(self.application['Name']))
                self.logger.info('Name: {0}, Path: {1}'.format(self.application['Name'], self.application['InstallPath']))
            else:
                self.logger.info('Application TheSkyX not found on computer')

    def start(self):
        pass

    def stop(self):
        pass

    def getStatus(self):
        captureResponse = {}
        try:
            command = '/* Java Script */'
            command += 'var Out = "";'
            command += 'ccdsoftCamera.Asynchronous=0;'
            command += 'Out=String(\'{"ExposureStatus":"\'+ccdsoftCamera.ExposureStatus+\'","Status":"\'+ccdsoftCamera.Status+\'"}\');'
            # command += 'Out=ccdsoftCamera.ExposureStatus;'
            success, response = self.sendCommand(command)
            captureResponse = json.loads(response)
        except Exception as e:
            self.logger.error('Exception by getting status, error {0}'.format(e))
            success = False
            captureResponse['ExposureStatus'] = ''
            captureResponse['Status'] = ''
        finally:
            pass

        if success:
            if captureResponse['ExposureStatus'].startswith('Exposing Light'):
                captureResponse['ExposureStatus'] = 'Exposing Light'
            if captureResponse['ExposureStatus'] in self.CAMERASTATUS:
                self.data['Camera']['Status'] = self.CAMERASTATUS[captureResponse['ExposureStatus']]
                if self.CAMERASTATUS[captureResponse['ExposureStatus']] == 'DISCONNECTED':
                    self.data['Camera']['CONNECTION']['CONNECT'] = 'Off'
                    self.data['Solver']['CONNECTION']['CONNECT'] = 'Off'
                else:
                    self.data['Camera']['CONNECTION']['CONNECT'] = 'On'
                    self.data['Solver']['CONNECTION']['CONNECT'] = 'On'
            else:
                self.logger.error('Unknown camera status: {0}'.format(captureResponse['ExposureStatus']))
        else:
            self.data['Camera']['Status'] = 'ERROR'
            self.data['Camera']['CONNECTION']['CONNECT'] = 'Off'
            self.data['Solver']['Status'] = 'ERROR'
            self.data['Solver']['CONNECTION']['CONNECT'] = 'Off'

        self.cameraStatusText.emit(self.data['Camera']['Status'])
        # construct exposure time if present
        if '(' in captureResponse['Status']:
            exposeval = float(captureResponse['Status'].split('(', 1)[1].split('Left)')[0].replace(',', '.'))
            self.cameraExposureTime.emit('{0:02.0f}'.format(exposeval))
        else:
            self.cameraExposureTime.emit('---')

        if 'CONNECTION' in self.data['Camera']:
            if self.data['Camera']['CONNECTION']['CONNECT'] == 'On':
                self.app.workerModelingDispatcher.signalStatusCamera.emit(3)
            else:
                self.app.workerModelingDispatcher.signalStatusCamera.emit(2)
        else:
            self.app.workerModelingDispatcher.signalStatusCamera.emit(0)

        if 'CONNECTION' in self.data['Solver']:
            if self.data['Solver']['CONNECTION']['CONNECT'] == 'On':
                self.app.workerModelingDispatcher.signalStatusSolver.emit(3)
            else:
                self.app.workerModelingDispatcher.signalStatusSolver.emit(2)
        else:
            self.app.workerModelingDispatcher.signalStatusSolver.emit(0)

    def getCameraProps(self):
        if 'CONNECTION' in self.data['Camera']:
            if self.data['Camera']['CONNECTION']['CONNECT'] == 'On':
                # TODO: Get the chance to implement subframe on / off if a camera doesn't support this
                # TODO: Gain setting in CMOS Cameras necessary ?
                try:
                    command = '/* Java Script */'
                    command += 'var Out = "";'
                    command += 'ccdsoftCamera.Asynchronous=0;'
                    command += 'Out=String(\'{"WidthInPixels":"\'+ccdsoftCamera.WidthInPixels+\'","HeightInPixels":"\'+ccdsoftCamera.HeightInPixels+\'"}\');'
                    success, response = self.sendCommand(command)
                    captureResponse = json.loads(response)
                except Exception as e:
                    self.logger.error('Exception by getting props, error {0}'.format(e))
                    return False, 'Request failed', '', '', '', ''

                if success:
                    self.data['Camera']['Gain'] = ['High']
                    self.data['Camera']['Message'] = ''
                    if False:
                        self.data['Camera']['CCD_FRAME'] = {}
                        self.data['Camera']['CCD_FRAME']['HEIGHT'] = 0
                        self.data['Camera']['CCD_FRAME']['WIDTH'] = 0
                        self.data['Camera']['CCD_FRAME']['X'] = 0
                        self.data['Camera']['CCD_FRAME']['Y'] = 0
                    self.data['Camera']['CCD_INFO'] = {}
                    self.data['Camera']['CCD_INFO']['CCD_MAX_X'] = int(captureResponse['WidthInPixels'])
                    self.data['Camera']['CCD_INFO']['CCD_MAX_Y'] = int(captureResponse['HeightInPixels'])

        if self.isRunning:
            PyQt5.QtCore.QTimer.singleShot(self.CYCLEPROPS, self.setCameraProps)

    def sendCommand(self, command):
        try:
            tsxSocket = socket.socket()
            tsxSocket.connect((self.host, self.port))
            tsxSocket.send(command.encode())
            response = str(tsxSocket.recv(1024).decode())

            if response.endswith(self.responseSuccess):
                response = response.replace(self.responseSuccess, '')
                return True, response
            else:
                return False, response
        except Exception as e:
            self.logger.error('error: {0}'.format(e))
            return False, format(e)
        finally:
            # noinspection PyUnboundLocalVariable
            tsxSocket.close()

    def getImage(self, imageParams):
        # TODO: how is TSX dealing with ISO settings for DSLR?
        try:
            command = '/* Java Script */'
            command += 'ccdsoftCamera.Asynchronous=1;'
            if imageParams['CanSubframe']:
                command += 'ccdsoftCamera.Subframe=1;'
                command += 'ccdsoftCamera.SubframeLeft=' + str(imageParams['OffX']) + ';'
                command += 'ccdsoftCamera.SubframeTop=' + str(imageParams['OffY']) + ';'
                command += 'ccdsoftCamera.SubframeRight=' + str(imageParams['OffX'] + imageParams['SizeX']) + ';'
                command += 'ccdsoftCamera.SubframeBottom=' + str(imageParams['OffY'] + imageParams['SizeY']) + ';'
            else:
                command += 'ccdsoftCamera.Subframe=0;'

            if imageParams['Speed'] == 'HiSpeed':
                command += 'ccdsoftCamera.setPropStr("m_csExCameraMode", "Fast Image Download");'
            else:
                command += 'ccdsoftCamera.setPropStr("m_csExCameraMode", "High Image Quality");'

            command += 'ccdsoftCamera.BinX='+str(imageParams['Binning'])+';'
            command += 'ccdsoftCamera.BinY='+str(imageParams['Binning'])+';'
            command += 'ccdsoftCamera.ExposureTime='+str(imageParams['Exposure'])+';'
            command += 'ccdsoftCamera.AutoSavePath="'+str(imageParams['BaseDirImages'])+'";'
            command += 'ccdsoftCamera.AutoSaveOn=1;'
            command += 'ccdsoftCamera.Frame="cdLight";'
            command += 'ccdsoftCamera.TakeImage();'
            success, response = self.sendCommand(command)
            if success:
                command = '/* Java Script */'
                command += 'ccdsoftCamera.Asynchronous=1;'
                command += 'var Out = "";'
                command += 'Out=ccdsoftCamera.IsExposureComplete;'
                while True:
                    success, response = self.sendCommand(command)
                    if response == '1':
                        break
                    else:
                        time.sleep(0.2)
                        PyQt5.QtWidgets.QApplication.processEvents()
            command = '/* Java Script */'
            command += 'ccdsoftCamera.Asynchronous=1;'
            command += 'var Out = "";'
            command += 'Out=ccdsoftCamera.LastImageFileName;'
            success, response = self.sendCommand(command)
            imageParams['Message'] = 'OK'
            imageParams['Imagepath'] = response
        except Exception as e:
            self.logger.error('error: {0}'.format(e))
            imageParams['Message'] = 'Request failed'
            imageParams['Imagepath'] = ''
        finally:
            return imageParams

    def connectCamera(self):
        if self.isRunning:
            tsxSocket = None
            try:
                tsxSocket = socket.socket()
                tsxSocket.connect((self.host, self.port))
                command = '/* Java Script */ '
                command += 'ccdsoftCamera.Connect();'
            except Exception as e:
                self.logger.error('error: {0}'.format(e))
            finally:
                if tsxSocket:
                    tsxSocket.close()

    def disconnectCamera(self):
        if self.isRunning:
            tsxSocket = None
            try:
                tsxSocket = socket.socket()
                tsxSocket.connect((self.host, self.port))
                command = '/* Java Script */ '
                command += 'ccdsoftCamera.Disconnect();'
                self.sendCommand(command)
            except Exception as e:
                self.logger.error('error: {0}'.format(e))
            finally:
                if tsxSocket:
                    tsxSocket.close()
